# General imports

from AccessControl import ClassSecurityInfo
from Globals import InitializeClass
from types import FunctionType as function
from zope.component import adapts
from zope.interface import Interface, implements, alsoProvides, providedBy
from zope.app.annotation.interfaces import IAnnotations
from persistent.dict import PersistentDict
from Products.CMFCore.permissions import ManageProperties
from Products.Archetypes.Storage.annotation import AnnotationStorage
from Products.Archetypes.ClassGen import Generator
from Products.Archetypes.utils import _getSecurity
from Products.Archetypes.Renderer import renderer
from Products.Archetypes.BaseObject import BaseObject

from interfaces import IAnnotatedMetadata, IAnnotatableMetadata, IRemoteSchema

# local imports
from config import ANNO_KEY

class AnnotatedMetadata(object):
    """adapts an object to support our metadata storage
    """
    implements(IAnnotatedMetadata)
    adapts(IAnnotatableMetadata)

    _anno_key = ANNO_KEY
    _guarded_writes = True

    security = ClassSecurityInfo()
    security.setDefaultAccess('allow')

    def __init__(self, context):
        """ init the adapter """
        self.context = context
        annotations = IAnnotations(self.context)
        self._metadata = annotations.get(self._anno_key, None)
        if self._metadata is None:
            self._metadata = PersistentDict()
            annotations[self._anno_key] = self._metadata


    security.declarePublic('test')
    def test(self, nace=None):
        """ tester """
        kw = {'nace': nace}
        annotations = IAnnotations(self.context)
        self._metadata = annotations.get(self._anno_key, None)
        if 'nace' in kw.keys():
            self._metadata['nace'] = kw['nace']
        annotations[self._anno_key] = self._metadata


    security.declarePublic('setMetadataValues')
    def setValues(self, vals):
        """Sets the values given for existing metadata definitions
        """
        annotations = IAnnotations(self.context)
        self._metadata = annotations.get(self._anno_key, None)
        for key in vals.keys():
            self._metadata[key] = vals[key]
        annotations[self._anno_key] = self._metadata


    security.declarePublic('getValue')
    def getValue(self, name, default=None):
        """Gets a value from the set
        """
        annotations = IAnnotations(self.context)
        self._metadata = annotations.get(self._anno_key)
        return self._metadata.get(name, default)

    security.declarePublic('getKeys')
    def getKeys(self):
        """Retrieves all available keys in the annotation storage
        """
        annotations = IAnnotations(self.context)
        self._metadata = annotations.get(self._anno_key, None)
        return self._metadata.keys()


InitializeClass(AnnotatedMetadata)


class RemoteSchema(object):
    """adapts an object to support annotation metadata storage. The Schema is
       taken from the portal_annotatedmetadata tool and the values are taken from the annotations on
       the current context
    """

    implements(IRemoteSchema)
    _guarded_writes = True
    security = ClassSecurityInfo()
    security.setDefaultAccess('allow')

    def __init__(self, context):
        """ init the adapter """
        self.context = context

        # currently hardcoded to use the helper object METADATASET which is editable by ATSE and stores our remote schema
        self.remoteob = context.portal_url.getPortalObject().portal_annotatablemetadata.METADATASET
        security = _getSecurity(self.__class__)

        # I want to use the Generator to generate my method names
        G = Generator()
        # We need to make sure that the autogenerated Accessors from the remoteob are available on our adapter
        # and they need to access the annotation storage
        S = self.remoteob.Schema()
        for key in S.keys():
            field = S[key]
            # XXX: I have not yet fully understood why there also is an accessor called _v_name_accessor which does the same as getName... as an intermediate solution I generate both. Ugly, I hope to understand later.

            # Generating getName
            acc = G.computeMethodName(field, 'r')
            self.makeMethod(field, acc)
            security.declarePublic(acc)

            # Generating _v_name_accessor
            vacc = field.accessor
            self.makeMethod(field, vacc)
            security.declarePublic(vacc)

        InitializeClass(self.__class__)

    def makeMethod(self, field, methodName):
        """ Taken from Archetypes.ClassGen.py to generate accessors according
            to the remote schema which access the annotation storage.
        """
        name = field.getName()
        method = None
        def generatedAccessor():
            """Default Annotated Accessor."""
            annotations = IAnnotations(self.context)
            md = annotations.get(ANNO_KEY, {})
            return md.get(name, None)

        method = generatedAccessor

        # Zope security requires all security protected methods to have a
        # function name. It uses this name to determine which roles are allowed
        # to access the method.
        # This code is renaming the internal name from e.g. generatedAccessor to
        # methodName.
        method = function(method.func_code,
                          method.func_globals,
                          methodName,
                          method.func_defaults,
                          method.func_closure,
                         )
        setattr(self, methodName, method)

    def Schema(self):
        """ Returns the remote Schema. This is needed to get the schema from the remote object instead of the local one
        """
        return self.remoteob.Schema()

    security.declarePublic('Schemata')
    def Schemata(self):
        """ Returns the remote Schemata.  This is needed to get the schemata from the remote object instead of the local one
        """
        return self.remoteob.Schemata()

    security.declarePublic('fields')
    def fields(self):
        """ fetches the fields from the remote Schema.
        """
        schema = self.Schema()
        fields = schema.fields()
        return fields

    security.declarePublic('widget')
    def widget(self, field_name, mode="view", field=None, **kwargs):
        """Returns the rendered widget. Taken from BaseObject.widget because I want to change the field and the accessor
           Basically this method displays the widget based on the remote schema using values from the annotation storage on the current context by using autogenerated accessors.
           The field is taken from the schema of the remote object
           The accessor is taken from this adapter and is autogenerated in the __init__ method.
        """
        # Get the field from the remote object
        if field is None:
            S = self.Schema()
            field = S._fields.get(field_name)
        widget = field.widget

        # determine the name of the autogenerated accessor method and fetch it
        G = Generator()
        accName = G.computeMethodName(field, 'r')
        acc = getattr(self, accName)
        # return the rendered widget on the current context object with the remote field but the modified accessor.
        return renderer.render(field_name, mode, widget, self.context, field=field,
                               accessor=acc, **kwargs)


InitializeClass(RemoteSchema)
